# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging

from flask import render_template

from datacube_ows.data import get_map
from datacube_ows.feature_info import feature_info
from datacube_ows.http_utils import cache_control_headers, get_service_base_url
from datacube_ows.ogc_exceptions import WMSException, WMTSException
from datacube_ows.ows_configuration import get_config
from datacube_ows.utils import log_call

_LOG = logging.getLogger(__name__)




# NB. No need to disambiguate method names shared with WMS because WMTS requires
# a "SERVICE" parameter with every request.

@log_call
def handle_wmts(nocase_args) -> tuple:
    operation = nocase_args.get("request", "").upper()
    # WMS operation Map
    if not operation:
        raise WMTSException("No operation specified", locator="Request parameter")
    elif operation == "GETCAPABILITIES":
        return get_capabilities(nocase_args)
    elif operation == "GETTILE":
        return get_tile(nocase_args)
    elif operation == "GETFEATUREINFO":
        return get_feature_info(nocase_args)
    else:
        raise WMTSException("Unrecognised operation: %s" % operation, WMTSException.OPERATION_NOT_SUPPORTED,
                           "Request parameter")


@log_call
def get_capabilities(args) -> tuple:
    # TODO: Handle updatesequence request parameter for cache consistency.
    # Note: Only WMS v1.0.0 exists at this stage, so no version negotiation is necessary
    # Extract layer metadata from Datacube.
    cfg = get_config()
    url = args.get('Host', args['url_root'])
    base_url = get_service_base_url(cfg.allowed_urls, url)
    section = args.get("section")
    if section:
        section = section.lower()
    show_service_id = False
    show_service_provider = False
    show_ops_metadata = False
    show_contents = False
    show_themes = False
    if section is None:
        show_service_id = True
        show_service_provider = True
        show_ops_metadata = True
        show_contents = True
        show_themes = True
    else:
        sections = section.split(",")
        for s in sections:
            if s == "all":
                show_service_id = True
                show_service_provider = True
                show_ops_metadata = True
                show_contents = True
                show_themes = True
            elif s == "serviceidentification":
                show_service_id = True
            elif s == "serviceprovider":
                show_service_provider = True
            elif s == "operationsmetadata":
                show_ops_metadata = True
            elif s == "contents":
                show_contents = True
            elif s == "themes":
                show_themes = True
            else:
                raise WMTSException("Invalid section: %s" % section,
                                WMTSException.INVALID_PARAMETER_VALUE,
                                locator="Section parameter")
    headers = cache_control_headers(cfg.wms_cap_cache_age)
    headers["Content-Type"] = "application/xml"
    return (
        render_template(
            "wmts_capabilities.xml",
            cfg=cfg,
            base_url=base_url,
            show_service_id=show_service_id,
            show_service_provider=show_service_provider,
            show_ops_metadata=show_ops_metadata,
            show_contents=show_contents,
            show_themes=show_themes),
        200,
        cfg.response_headers(headers)
    )


@log_call
def wmts_args_to_wms(args, cfg) -> dict:
    layer = args.get("layer")
    style = args.get("style")
    format_ = args.get("format")
    time = args.get("time", "")
    tileMatrixSet = args.get("tilematrixset")
    tileMatrix = args.get("tilematrix")
    row = args.get("tilerow")
    col = args.get("tilecol")

    wms_args = {
        "version": "1.3.0",
        "service": "WMS",
        "request": "GetMap",
        "styles": style,
        "layers": layer,
        "time": time,
        "width": 256,
        "height": 256,
        "format": format_,
        "exceptions": "application/vnd.ogc.se_xml",
        "requestid": args["requestid"]
    }

    tms = cfg.tile_matrix_sets.get(tileMatrixSet)
    if not tms:
        for _tms in cfg.tile_matrix_sets.values():
            if tileMatrixSet == _tms.wkss:
                tms = _tms
                break

    if tms is None:
        raise WMTSException("Invalid Tile Matrix Set: " + tileMatrixSet)

    wms_args["crs"] = tms.crs_name
    _ = cfg.published_CRSs[tms.crs_name]
    try:
        tileMatrix = int(tileMatrix)
        if tileMatrix < 0 or tileMatrix >= len(tms.scale_set):
            raise WMTSException(f"Invalid Tile Matrix: {tileMatrix}")
    except ValueError:
        raise WMTSException(f"Invalid Tile Matrix: {tileMatrix}")
    try:
        row = int(row)
    except ValueError:
        raise WMTSException(f"Invalid Tile Row {row}")
    try:
        col = int(col)
    except ValueError:
        raise WMTSException(f"Invalid Tile Col: {col}")
    wms_args["bbox"] = "%f,%f,%f,%f" % tms.wms_bbox_coords(tileMatrix, row, col)

    # GetFeatureInfo only args
    if "i" in args:
        wms_args["i"] = args["i"]
        wms_args["j"] = args.get("j", "")
        wms_args["info_format"] = args.get("infoformat", "")


    if args.get("ows_stats"):
        wms_args["ows_stats"] = "y"

    return wms_args


@log_call
def get_tile(args) -> tuple:
    cfg = get_config()
    wms_args = wmts_args_to_wms(args, cfg)

    try:
        return get_map(wms_args)
    except WMSException as wmse:
        first_error = wmse.errors[0]
        e = WMTSException(first_error["msg"],
                            code=first_error["code"],
                            locator=first_error["locator"],
                            http_response=wmse.http_response)
        for error in wmse.errors[1:]:
            e.add_error(error["msg"], code=error["code"], locator=error["locator"])
        raise e


@log_call
def get_feature_info(args) -> tuple:
    cfg = get_config()
    wms_args = wmts_args_to_wms(args, cfg)
    wms_args["query_layers"] = wms_args["layers"]

    try:
        return feature_info(wms_args)
    except WMSException as wmse:
        first_error = wmse.errors[0]
        e = WMTSException(first_error["msg"],
                          code=first_error["code"],
                          locator=first_error["locator"],
                          http_response=wmse.http_response)
        for error in wmse.errors[1:]:
            e.add_error(error["msg"], code=error["code"], locator=error["locator"])
        raise e



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from affine import Affine
from odc.geo import geom as odc_geom
from typing_extensions import override


class WCSScalerException(Exception):
    pass


class WCSScalerUnknownDimension(WCSScalerException):
    def __init__(self, dim, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.dim = dim


class WCSScalerOverspecifiedDimension(WCSScalerException):
    pass


class WCSScalarIllegalSize(WCSScalerException):
    pass


class SpatialParameter:
    def __init__(self, layer, crs, x=None, y=None) -> None:
        self.layer = layer
        self.crs_def = self.layer.global_cfg.published_CRSs[crs]
        self.x = x
        self.y = y

    def is_x_dim(self, dimension) -> bool:
        if dimension == self.crs_def['horizontal_coord'].lower():
            return True
        elif dimension == self.crs_def['vertical_coord'].lower():
            return False
        elif dimension == self.layer.native_CRS_def['horizontal_coord'].lower():
            return True
        elif dimension == self.layer.native_CRS_def['vertical_coord'].lower():
            return False
        elif dimension in ("x", "i", "lon", "long", "lng", "longitude"):
            return True
        elif dimension in ("y", "j", "lat", "latitude"):
            return False
        else:
            raise WCSScalerUnknownDimension(dimension)

    def __getitem__(self, dim):
        if self.is_x_dim(dim):
            return self.x
        else:
            return self.y

    def __setitem__(self, dim, val) -> None:
        if self.is_x_dim(dim):
            self.x = val
        else:
            self.y = val

    def __getattr__(self, dim):
        return self[dim]

    @override
    def __setattr__(self, dim, val) -> None:
        if dim in ("x", "y", "layer", "crs_def"):
            super().__setattr__(dim, val)
        else:
            try:
                self[dim] = val
            except WCSScalerUnknownDimension:
                super().__setattr__(dim, val)

    def set(self, x, y) -> None:
        self.x = x
        self.y = y


class WCSScaler:
    def __init__(self, layer, crs=None) -> None:
        self.layer = layer
        self.cfg = self.layer.global_cfg
        if crs:
            self.crs = crs
        else:
            self.crs = self.layer.native_CRS
        self.min = SpatialParameter(self.layer, self.crs)
        self.max = SpatialParameter(self.layer, self.crs)
        self.size = SpatialParameter(self.layer, self.crs)
        self.subsetted = SpatialParameter(self.layer, self.crs, False, False)

    @property
    def crs(self):
        return self._crs

    @crs.setter
    def crs(self, crs) -> None:
        self.crs_def = self.layer.global_cfg.published_CRSs[crs]
        self._crs = crs

    def set_size(self, dim, size) -> None:
        if size <= 0:
            raise WCSScalarIllegalSize()
        if isinstance(size, float):
            size = int(size + 0.5)
        if self.size[dim] is None:
            self.size[dim] = size
        else:
            raise WCSScalerOverspecifiedDimension()

    def slice(self, dimension, value) -> None:
        self.min[dimension] = value
        self.max[dimension] = value
        self.subsetted[dimension] = True

    def is_slice(self, dim) -> bool:
        return self.subsetted[dim] and self.min[dim] == self.max[dim]

    def dim(self, dim) -> tuple:
        return self.size[dim], self.min[dim], self.max[dim]

    def trim(self, dimension, lower, higher) -> None:
        self.min[dimension] = lower
        self.max[dimension] = higher
        self.subsetted[dimension] = True

    def to_crs(self, new_crs):
        grid = self.layer.grids[new_crs]
        skip_x_xform = False
        skip_y_xform = False
        if not self.subsetted.x and not self.subsetted.y:
            # Neither axis subsetted
            self.min.x = self.layer.ranges.bboxes[new_crs]["left"]
            self.max.x = self.layer.ranges.bboxes[new_crs]["right"]
            self.min.y = self.layer.ranges.bboxes[new_crs]["bottom"]
            self.max.y = self.layer.ranges.bboxes[new_crs]["top"]
            self.crs = new_crs
        elif not self.subsetted.x or not self.subsetted.y:
            # One axis subsetted
            if self.subsetted.x:
                self.min.y = self.layer.ranges.bboxes[self.crs]["bottom"]
                self.max.y = self.layer.ranges.bboxes[self.crs]["top"]
                skip_y_xform = True
            if self.subsetted.y:
                self.min.x = self.layer.ranges.bboxes[self.crs]["left"]
                self.max.x = self.layer.ranges.bboxes[self.crs]["right"]
                skip_x_xform = True
        else:
            # Both axes subsetted
            pass

        if self.crs != new_crs:
            is_point = False
            # Prepare geometry for transformation
            old_crs_obj = self.cfg.crs(self.crs)
            if self.is_slice("x") and self.is_slice("y"):
                geom = odc_geom.point(self.min.x, self.min.y, old_crs_obj)
                is_point = True
            elif self.is_slice("x") or self.is_slice("y"):
                geom = odc_geom.line(
                    [
                        (self.min.x, self.min.y),
                        (self.max.x, self.max.y)
                    ], old_crs_obj)
            else:
                geom = odc_geom.polygon(
                    [
                        (self.min.x, self.min.y),
                        (self.min.x, self.max.y),
                        (self.max.x, self.max.y),
                        (self.max.x, self.min.y),
                        (self.min.x, self.min.y),
                    ],
                    old_crs_obj
                )
            new_crs_obj = self.cfg.crs(new_crs)
            grid = self.layer.grids[new_crs]
            if is_point:
                prj_pt = geom.to_crs(new_crs_obj)
                x, y = prj_pt.coords[0]
                self.min.set(x, y)
                self.max.set(x + grid["resolution"][0],
                             y + grid["resolution"][1])
                self.size.set(1, 1)
            else:
                proj_geom = geom.to_crs(new_crs_obj)
                bbox = proj_geom.boundingbox
                if skip_x_xform:
                    self.min.x = self.layer.ranges.bboxes[new_crs]["left"]
                    self.max.x = self.layer.ranges.bboxes[new_crs]["right"]
                else:
                    self.min.x = bbox.left
                    self.max.x = bbox.right
                if skip_y_xform:
                    self.min.y = self.layer.ranges.bboxes[new_crs]["bottom"]
                    self.max.y = self.layer.ranges.bboxes[new_crs]["top"]
                else:
                    self.min.y = bbox.bottom
                    self.max.y = bbox.top

                self.quantise_to_resolution(grid)
            self.crs = new_crs
        else:
            self.quantise_to_resolution(grid)

    def quantise_to_resolution(self, grid) -> None:
        for idx, dim in enumerate("xy"):
            if abs(self.max[dim] - self.min[dim]) < abs(grid["resolution"][idx] * 1.5):
                self.max[dim] = self.min[dim] + grid["resolution"][idx]
                self.size[dim] = 1

    def scale_axis(self, dimension, factor) -> None:
        dim_size, dim_min, dim_max = self.dim(dimension)
        if dim_size is not None:
            raise WCSScalerOverspecifiedDimension()
        grid = self.layer.grids[self.crs]
        if self.min.is_x_dim(dimension):
            res = grid["resolution"][0]
        else:
            res = grid["resolution"][1]
        scaled_size = abs(
            (dim_max - dim_min) * factor / res
        )
        self.set_size(dimension, scaled_size)

    def scale_size(self, dimension, size) -> None:
        self.set_size(dimension, size)

    def scale_extent(self, dimension, low, high) -> None:
        # TODO: What is this actually supposed to mean?
        self.set_size(dimension, high - low)

    def affine(self) -> Affine:
        if self.size.x is None:
            self.scale_axis("x", 1.0)
        if self.size.y is None:
            self.scale_axis("y", 1.0)

        x_scale = (self.max.x - self.min.x) / self.size.x
        # Y axis is reversed: image coordinate conventions
        y_scale = (self.min.y - self.max.y) / self.size.y
        # if self.crs_def["vertical_coord_first"]:
            # This should probably happen, but can't because PostGIS wants
            # coords to be horizontal first, regardless of what the CRS says.
            # trans_aff = Affine.translation(self.min.y, self.max.x)
            # scale_aff = Affine.scale(y_scale, x_scale)
        # else:
        trans_aff = Affine.translation(self.min.x, self.max.y)
        scale_aff = Affine.scale(x_scale, y_scale)
        return trans_aff * scale_aff



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import io
import logging
from collections import defaultdict
from dataclasses import dataclass
from decimal import ROUND_HALF_UP, Decimal
from math import isclose
from typing import Any, Union, cast
from typing_extensions import override
from collections.abc import Hashable, Iterable, MutableMapping

import matplotlib
import numpy
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, to_hex, to_rgba
from numpy import ubyte
from xarray import DataArray, Dataset

from datacube_ows.config_utils import (CFG_DICT, ConfigException,
                                       FunctionWrapper, OWSMetadataConfig)
from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.expression import Expression

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.ows_configuration import OWSNamedLayer

_LOG = logging.getLogger(__name__)

@dataclass
class RampNode:
    value: int | float
    color: str
    alpha: int | float | None = None

    @property
    def rgba(self):
        return to_rgba(self.color, self.alpha)

    def with_value(self, value: int | float) -> "RampNode":
        return RampNode(value, self.color, self.alpha)


RAMP_SPEC = list[CFG_DICT]
RampRepr = list[RampNode]


def make_ramp_representation(ramp_spec: RAMP_SPEC, style_name: str) -> RampRepr:
    rep: RampRepr = []
    for node in ramp_spec:
        if "value" not in node:
            raise ConfigException(f'Color ramp element without a value in style {style_name}')
        if "color" not in node:
            raise ConfigException(f'Color ramp element without a color in style {style_name}')
        if "legend" in node:
                raise ConfigException(
                    f"Style {style_name} uses a no-longer supported format for legend configuration.  " +
                    "Please refer to the documentation and update your config")
        rep.append(
            RampNode(
                float(cast(float | str | int, node["value"])),
                cast(str, node["color"]),
                alpha=None if node.get("alpha") is None else float(cast(str | int | float, node["alpha"]))
            )
        )
    return rep


UNSCALED_DEFAULT_RAMP = [
        RampNode(-1e-24, "#000080", alpha=0.0),
        RampNode(0.0, "#000080"),
        RampNode(0.1, "#0000FF"),
        RampNode(0.3, "#00FFFF"),
        RampNode(0.5, "#00FF00"),
        RampNode(0.7, "#FFFF00"),
        RampNode(0.9, "#FF0000"),
        RampNode(1.0, "#800000"),
]


def scale_unscaled_ramp(rmin: int | float | str, rmax: int | float | str, unscaled: RampRepr) -> RampRepr:
    """
    Take a unscaled (normalised) ramp that covers values from 0.0 to 1.0 and scale it linearly to cover the
    provided range.

    :param rmin: The new minimum value for the ramp range.
    :param rmax: The new maximum value for the ramp range.
    :param unscaled: The unscaled (normalised) ramp.
    :return: The scaled ramp.
    """
    if isinstance(rmin, float):
        nmin: float = rmin
    else:
        nmin = float(rmin)
    if isinstance(rmax, float):
        nmax: float = rmax
    else:
        nmax = float(rmax)
    return [
        u.with_value((nmax - nmin) * float(u.value) + nmin)
        for u in unscaled
    ]


def crack_ramp(ramp: RampRepr) -> tuple[
    list[float],
    list[float], list[float],
    list[float], list[float],
]:
    """
    Split a colour ramp into separate (input) value and (output) RGBA lists.

    :param ramp: input (scaled) colour-ramp definition
    :return: A tuple of four lists of floats: representing values, red, green, blue, alpha.
    """
    values = cast(list[float], [])
    red = cast(list[float], [])
    green = cast(list[float], [])
    blue = cast(list[float], [])
    alpha = cast(list[float], [])
    for r in ramp:
        values.append(float(r.value))
        cr, cg, cb, ca = r.rgba
        red.append(cr)
        green.append(cg)
        blue.append(cb)
        alpha.append(ca)

    return values, red, green, blue, alpha


def read_mpl_ramp(mpl_ramp: str) -> RampRepr:
    """
    Extract a named colour ramp from Matplotlib as a normalised OWS-compatible ramp specification

    :param mpl_ramp: The name of Matplotlib colour ramp
    :return: A normalised ramp specification.
    """
    unscaled_cmap: RampRepr = []
    try:
        cmap = plt.get_cmap(mpl_ramp)
    except:
        raise ConfigException(f"Invalid Matplotlib name: {mpl_ramp}")
    val_range = numpy.arange(0.1, 1.1, 0.1)
    rgba_hex = to_hex(cmap(0.0))
    unscaled_cmap.append(RampNode(0.0, rgba_hex))
    for val in val_range:
        rgba_hex = to_hex(cast(tuple[float, float, float, float], cmap(val)))
        unscaled_cmap.append(
            RampNode(float(val), rgba_hex)
        )
    return unscaled_cmap


class ColorRamp:
    """
    Represents a colour ramp for image and legend rendering purposes
    """
    def __init__(self, style: StyleDefBase,
                       ramp_cfg: CFG_DICT,
                       legend: "RampLegendBase") -> None:
        """
        :param style: The style owning the ramp
        :param ramp_cfg: Style config
        """
        self.style = style
        if "color_ramp" in ramp_cfg:
            raw_scaled_ramp = make_ramp_representation(cast(RAMP_SPEC, ramp_cfg["color_ramp"]), self.style.name)
        else:
            rmin, rmax = cast(list[float], ramp_cfg["range"])
            unscaled_ramp = UNSCALED_DEFAULT_RAMP
            if "mpl_ramp" in ramp_cfg:
                unscaled_ramp = read_mpl_ramp(cast(str, ramp_cfg["mpl_ramp"]))
            raw_scaled_ramp = scale_unscaled_ramp(rmin, rmax, unscaled_ramp)
        self.ramp = raw_scaled_ramp

        self.values = cast(list[float], [])
        self.components = cast(MutableMapping[str, list[float]], {})
        self.crack_ramp()

        # Handle the mutual interdepencies between the ramp and the legend
        # 1. Let legend read its defaults from this ramp if needed
        legend.register_ramp(self)
        # 2. Extend our colour ramp to support legend if needed
        if self.style.auto_legend:
            fleg_begin = float(legend.begin)
            fleg_end = float(legend.end)
            leg_begin_in_ramp = False
            leg_end_in_ramp = False
            leg_begin_before_idx = None
            leg_end_before_idx = None
            for idx, col_point in enumerate(self.ramp):
                col_val = col_point.value
                if not leg_begin_in_ramp and leg_begin_before_idx is None:
                    if isclose(col_val, fleg_begin, abs_tol=1e-9):
                        leg_begin_in_ramp = True
                    elif col_val > fleg_begin:
                        leg_begin_before_idx = idx
                if not leg_end_in_ramp and leg_end_before_idx is None:
                    if isclose(col_val, fleg_end, abs_tol=1e-9):
                        end_in_ramp = True
                    elif col_val > fleg_end:
                        end_before_idx = idx
            if not leg_begin_in_ramp:
                rgba = self.rgba_at(fleg_begin)
                begin_col_point = RampNode(fleg_begin, to_hex(rgba))
                if leg_begin_before_idx is None:
                    self.ramp.append(begin_col_point)
                else:
                    self.ramp.insert(leg_begin_before_idx, begin_col_point)
                if leg_end_before_idx is not None:
                    leg_end_before_idx += 1
            if not leg_end_in_ramp:
                rgba = self.rgba_at(fleg_end)
                end_col_point = RampNode(fleg_end, to_hex(rgba))
                if leg_end_before_idx is None:
                    self.ramp.append(end_col_point)
                else:
                    self.ramp.insert(leg_end_before_idx, end_col_point)
            if not leg_end_in_ramp or not leg_begin_in_ramp:
                self.crack_ramp()

    def crack_ramp(self) -> None:
        values, r, g, b, a = crack_ramp(self.ramp)
        self.values = values
        self.components = {
            "red": r,
            "green": g,
            "blue": b,
            "alpha": a
        }

    def get_value(self, data: float | DataArray, band: str) -> numpy.ndarray | float:
        return numpy.interp(data, self.values, self.components[band])

    def get_8bit_value(self, data: DataArray, band: str) -> numpy.ndarray:
        val = cast(numpy.ndarray, self.get_value(data, band))
        val = val * 255
        # Is there a way to stop this raising a runtime warning?
        return val.astype(ubyte)

    def apply(self, data: DataArray) -> Dataset:
        imgdata = cast(MutableMapping[Hashable, Any], {})
        for band in self.components:
            imgdata[band] = (data.dims, self.get_8bit_value(data, band))
        imgdataset = Dataset(imgdata, coords=data.coords)
        return imgdataset

    def rgba_at(self, val: float) -> tuple[float, float, float, float]:
        return (
            float(self.get_value(val, "red")),
            float(self.get_value(val, "green")),
            float(self.get_value(val, "blue")),
            float(self.get_value(val, "alpha"))
        )


class RampLegendBase(StyleDefBase.Legend, OWSMetadataConfig):
    METADATA_ABSTRACT: bool = False
    METADATA_LEGEND_UNITS: bool = True
    METADATA_TICK_LABELS: bool = True

    def __init__(self, style_or_mdh: Union["StyleDefBase", "StyleDefBase.Legend"], cfg: CFG_DICT) -> None:
        super().__init__(style_or_mdh, cfg)
        raw_cfg = cast(CFG_DICT, self._raw_cfg)
        # Range - defaults deferred until we have parsed the associated ramp
        if "begin" not in raw_cfg:
            self.begin = Decimal("nan")
        else:
            self.begin = Decimal(cast(str | float | int, raw_cfg["begin"]))
        if "end" not in raw_cfg:
            self.end = Decimal("nan")
        else:
            self.end = Decimal(cast(str | float | int, raw_cfg["end"]))

        # decimal_places, rounder
        def rounder_str(prec: int) -> str:
            rstr = "1"
            if prec == 0:
                return rstr
            rstr += "."
            for i in range(prec - 1):
                rstr += "0"
            rstr += "1"
            return rstr

        self.decimal_places = cast(int, raw_cfg.get("decimal_places", 1))
        if self.decimal_places < 0:
            raise ConfigException("decimal_places cannot be negative")
        self.rounder = Decimal(rounder_str(self.decimal_places))

        # Ticks - Non-explicit tick values deferred until we have parsed the associated ramp
        ticks_handled = False
        self.ticks_every: Decimal | None = None
        self.tick_count: int | None = None
        self.ticks: list[Decimal] = []
        if "ticks_every" in raw_cfg:
            if "tick_count" in raw_cfg:
                raise ConfigException("Cannot use tick count and ticks_every in the same legend")
            if "ticks" in raw_cfg:
                raise ConfigException("Cannot use ticks and ticks_every in the same legend")
            self.ticks_every = Decimal(cast(int | float | str, raw_cfg["ticks_every"]))
            if self.ticks_every.is_zero() or self.ticks_every.is_signed():
                raise ConfigException("ticks_every must be greater than zero")
            ticks_handled = True
        if "ticks" in raw_cfg:
            if "tick_count" in raw_cfg:
                raise ConfigException("Cannot use tick count and ticks in the same legend")
            self.ticks = [Decimal(t) for t in cast(list[str | int | float], raw_cfg["ticks"])]
            ticks_handled = True
        if not ticks_handled:
            self.tick_count = int(cast(str | int, raw_cfg.get("tick_count", 1)))
            if self.tick_count < 0:
                raise ConfigException("tick_count cannot be negative")
        # prepare for tick labels
        self.cfg_labels = cast(MutableMapping[str, MutableMapping[str, str]], raw_cfg.get("tick_labels", {}))
        defaults = self.cfg_labels.get("default", {})
        self.lbl_default_prefix = defaults.get("prefix", "")
        self.lbl_default_suffix = defaults.get("suffix", "")
        self.tick_labels: list[str] = []
        # handle matplotlib args
        self.strip_location = cast(tuple[float, float, float, float],
                                   tuple(cast(Iterable[float], raw_cfg.get("strip_location", [0.05, 0.5, 0.9, 0.15]))))
        # throw error on legacy syntax
        self.fail_legacy()

    def fail_legacy(self) -> None:
        if any(
                legent in cast(CFG_DICT, self._raw_cfg)
                for legent in ["major_ticks", "offset", "scale_by", "radix_point"]
        ):
            raise ConfigException(
                f"Style {self.style.name} uses a no-longer supported format for legend configuration.  " +
                "Please refer to the documentation and update your config")

    def register_ramp(self, ramp: ColorRamp) -> None:
        if self.begin.is_nan():
            for col_def in ramp.ramp:
                if isclose(col_def.rgba[-1], 1.0, abs_tol=1e-9):
                    self.begin = Decimal(col_def.value)
                    break
            if self.begin.is_nan():
                self.begin = Decimal(ramp.ramp[0].value)
        if self.end.is_nan():
            for col_def in reversed(ramp.ramp):
                if isclose(col_def.rgba[-1], 1.0, abs_tol=1e-9):
                    self.end = Decimal(col_def.value)
                    break
            if self.end.is_nan():
                self.end = Decimal(ramp.ramp[-1].value)
        for t in self.ticks:
            if t < self.begin or t > self.end:
                raise ConfigException("Explicit ticks must all be within legend begin/end range")
        if self.ticks_every is not None:
            tickval = self.begin
            while tickval < self.end:
                self.ticks.append(tickval)
                tickval += self.ticks_every
            self.ticks.append(self.end)
        elif self.tick_count is not None:
            if self.tick_count == 0:
                self.ticks.append(self.begin)
            else:
                delta = self.end - self.begin
                dcount = Decimal(self.tick_count)
                for i in range(0, self.tick_count + 1):
                    tickval = self.begin + (Decimal(i) / dcount) * delta
                    self.ticks.append(tickval.quantize(self.rounder, rounding=ROUND_HALF_UP))

        # handle tick labels
        for tick in self.ticks:
            label_cfg = self.cfg_labels.get(str(tick))
            if label_cfg:
                prefix = label_cfg.get("prefix", self.lbl_default_prefix)
                suffix = label_cfg.get("suffix", self.lbl_default_suffix)
                label = label_cfg.get("label", str(tick))
                self.tick_labels.append(prefix + label + suffix)
            else:
                self.tick_labels.append(
                    self.lbl_default_prefix + str(tick) + self.lbl_default_suffix
                )
        self.parse_metadata(cast(CFG_DICT, self._raw_cfg))


    def tick_label(self, tick: Decimal) -> str | None:
        try:
            tick_idx = self.ticks.index(tick)
            metaval = self.read_local_metadata(f"lbl_{tick}")
            if metaval:
                return metaval
            else:
                return self.tick_labels[tick_idx]
        except ValueError:
            _LOG.error("'%s' is a not a valid tick", tick)
            return None

    def create_cdict_ticks(self) -> tuple[
        MutableMapping[str, list[tuple[float, float, float]]],
        MutableMapping[float, str],
    ]:
        normalize_factor = float(self.end) - float(self.begin)
        cdict = cast(MutableMapping[str, list[tuple[float, float, float]]], dict())
        bands = cast(MutableMapping[str, list[tuple[float, float, float]]], defaultdict(list))
        started = False
        finished = False
        for index, ramp_node in enumerate(self.style_or_mdh.color_ramp.ramp):
            if finished:
                break

            value = ramp_node.value
            normalized = (value - float(self.begin)) / float(normalize_factor)

            if not started:
                if isclose(value, float(self.begin), abs_tol=1e-9):
                    started = True
                else:
                    continue
            if not finished:
                if isclose(value, float(self.end), abs_tol=1e-9):
                    finished = True

            for band, intensity in self.style_or_mdh.color_ramp.components.items():
                bands[band].append((normalized, intensity[index], intensity[index]))

        for band, blist in bands.items():
            cdict[band] = blist

        ticks = cast(MutableMapping[float, str], dict())
        for tick in self.ticks:
            value = float(tick)
            normalized = (value - float(self.begin)) / float(normalize_factor)
            ticks[normalized] = cast(str, self.tick_label(tick))

        return cdict, ticks

    def display_title(self) -> str:
        if self.units:
            return f"{self.title}({self.units})"
        else:
            return self.title

    def plot_name(self) -> str:
        return f"{self.style.product.name}_{self.style.name}"

    @override
    def render(self, bytesio: io.BytesIO) -> None:
        cdict, ticks = self.create_cdict_ticks()
        plt.rcdefaults()
        if self.mpl_rcparams:
            plt.rcParams.update(self.mpl_rcparams)
        fig = plt.figure(figsize=(self.width, self.height))
        ax = fig.add_axes(self.strip_location)
        custom_map = LinearSegmentedColormap(self.plot_name(), cdict)  # type: ignore[arg-type]
        color_bar = matplotlib.colorbar.ColorbarBase(
            ax,
            cmap=custom_map,
            orientation="horizontal")
        color_bar.set_ticks(list(ticks.keys()))
        color_bar.set_ticklabels(list(ticks.values()))
        color_bar.set_label(self.display_title())
        plt.savefig(bytesio, format='png')

    # For MetadataConfig
    @property
    @override
    def default_title(self) -> str | None:
        return self.style.title



class ColorRampDef(StyleDefBase):
    """
    Colour ramp Style subclass
    """
    auto_legend = True

    def __init__(self,
                 product: "OWSNamedLayer",
                 style_cfg: CFG_DICT,
                 stand_alone: bool = False,
                 defer_multi_date: bool = False,
                 user_defined: bool = False) -> None:
        """"
        Constructor - refer to StyleDefBase
        """
        super(ColorRampDef, self).__init__(product, style_cfg,
                           stand_alone=stand_alone, defer_multi_date=True, user_defined=user_defined)
        style_cfg = cast(CFG_DICT, self._raw_cfg)
        self.color_ramp = ColorRamp(self, style_cfg, cast(ColorRampDef.Legend, self.legend_cfg))
        self.include_in_feature_info = bool(style_cfg.get("include_in_feature_info", True))

        if "index_function" in style_cfg:
            self.index_function: FunctionWrapper | Expression = FunctionWrapper(self,
                                                  cast(CFG_DICT, style_cfg["index_function"]),
                                                  stand_alone=self.stand_alone)
            if not self.stand_alone:
                for band in cast(list[str], style_cfg["needed_bands"]):
                    self.raw_needed_bands.add(band)
        elif "index_expression" in style_cfg:
            self.index_function = Expression(self, cast(str, style_cfg["index_expression"]))
            for band in self.index_function.needed_bands:
                self.raw_needed_bands.add(band)
            if self.stand_alone:
                self.needed_bands = set(self.local_band(b) for b in self.raw_needed_bands)
                self.flag_bands = set()
        else:
            raise ConfigException("Index function is required for index and hybrid styles. Style %s in layer %s" % (
                self.name,
                self.product.name
            ))
        if not defer_multi_date:
            self.parse_multi_date(style_cfg)

    def apply_index(self, data: Dataset) -> DataArray:
        """
        Caclulate index value across data.

        :param data: Input dataset
        :return: Matching dataarray carrying the index value
        """
        index_data = self.index_function(data)
        data['index_function'] = (index_data.dims, index_data.data)
        return data["index_function"]

    @override
    def transform_single_date_data(self, data: Dataset) -> Dataset:
        """
        Apply style to raw data to make an RGBA image xarray (single time slice only)

        :param data: Raw data, all bands.
        :return: RGBA ubyte xarray
        """
        d = self.apply_index(data)
        return self.color_ramp.apply(d)

    class Legend(RampLegendBase):
        @override
        def plot_name(self) -> str:
            return f"{self.style.product.name}_{self.style.name}_{self.style_or_mdh.min_count}"

    class MultiDateHandler(StyleDefBase.MultiDateHandler):
        auto_legend = True

        def __init__(self, style: "ColorRampDef", cfg: CFG_DICT) -> None:
            """
            First stage initialisation

            :param style: The parent style object
            :param cfg: The multidate handler configuration
            """
            super().__init__(style, cfg)
            if self.animate:
                self.feature_info_label: str | None = None
                self.color_ramp = style.color_ramp
                self.pass_raw_data = False
            else:
                self.feature_info_label = cast(str | None, cfg.get("feature_info_label", None))
                self.color_ramp = ColorRamp(style, cfg, cast(ColorRampDef.Legend, self.legend_cfg))
                self.pass_raw_data = bool(cfg.get("pass_raw_data", False))

        @override
        def transform_data(self, data: Dataset) -> Dataset:
            """
            Apply image transformation

            :param data: Raw data
            :return: RGBA image xarray.  May have a time dimension
            """
            if self.pass_raw_data:
                assert self.aggregator is not None  # For type-checker
                agg = self.aggregator(data)
            else:
                xformed_data = cast("ColorRampDef", self.style).apply_index(data)
                agg = cast(FunctionWrapper, self.aggregator)(xformed_data)
            return self.color_ramp.apply(agg)

        class Legend(RampLegendBase):
            pass


# Register ColorRampDef as Style subclass.
StyleDefBase.register_subclass(ColorRampDef, ("range", "color_ramp"))



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import warnings
from logging import Logger

from botocore.credentials import RefreshableCredentials
from datacube.utils.aws import configure_s3_access
from flask import Flask, request
from rasterio.errors import NotGeoreferencedWarning

from datacube_ows.ows_configuration import OWSConfig, get_config

__all__ = [
    'initialise_babel',
    'initialise_logger',
    'initialise_ignorable_warnings',
    'initialise_debugging',
    'initialise_sentry',
    'initialise_aws_credentials',
    'parse_config_file',
    'initialise_flask',
    'initialise_prometheus',
    'CredentialManager',
]

def initialise_logger(name: str | None = None) -> Logger:
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s'))
    _LOG = logging.getLogger(name)
    _LOG.addHandler(handler)
    # If invoked using Gunicorn, link our root logger to the gunicorn logger
    # this will mean the root logs will be captured and managed by the gunicorn logger
    # allowing you to set the gunicorn log directories and levels for logs
    # produced by this application
    _LOG.setLevel(logging.getLogger('gunicorn.error').getEffectiveLevel())
    return _LOG


def initialise_ignorable_warnings() -> None:
    # Suppress annoying rasterio warning message every time we write to a non-georeferenced image format
    warnings.simplefilter("ignore", category=NotGeoreferencedWarning)


def initialise_debugging(log: Logger | None = None) -> None:
    # PYCHARM Debugging
    if os.environ.get("PYDEV_DEBUG"):
        if os.environ["PYDEV_DEBUG"].lower() not in ("no", "false", "f", "n"):
            import pydevd_pycharm
            pydevd_pycharm.settrace('172.17.0.1', port=12321, stdoutToServer=True, stderrToServer=True)
            if log:
                log.info("PyCharm Debugging enabled")

def before_send(event, hint) -> None:
    if 'exc_info' in hint:
        exc_type, exc_value, tb = hint['exc_info']
        if isinstance(exc_value, AttributeError) and "object has no attribute 'GEOSGeom_destroy'" in str(exc_value):
            return None
    return event

def initialise_sentry(log: Logger | None = None) -> None:
    if os.environ.get("SENTRY_DSN"):
        import sentry_sdk
        from sentry_sdk.integrations.flask import FlaskIntegration
        SENTRY_ENV_TAG = os.environ.get("SENTRY_ENV_TAG") if os.environ.get("SENTRY_ENV_TAG") else "dev"
        sentry_sdk.init(
            dsn=os.environ["SENTRY_DSN"],
            environment=SENTRY_ENV_TAG,
            integrations = [FlaskIntegration()],
            before_send=before_send,
        )
        if log:
            log.info("Sentry initialised")

class CredentialManager:
    _instance = None

    def __new__(cls, log: Logger | None = None) -> "CredentialManager":
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, log: Logger | None = None) -> None:
        # Startup initialisation of libraries controlled by environment variables
        self.use_aws = False
        self.unsigned = False
        self.requester_pays = False
        self.credentials = None
        self.log = log

        if log:
            log.debug("Initialising CredentialManager")

        # Boto3/AWS
        if os.environ.get("AWS_DEFAULT_REGION") or os.environ.get("AWS_REGION"):
            if "AWS_NO_SIGN_REQUEST" in os.environ:
                env_nosign = os.environ["AWS_NO_SIGN_REQUEST"]
                if env_nosign.lower() in ("y", "t", "yes", "true", "1"):
                    unsigned = True
                    # Workaround for rasterio bug
                    os.environ["AWS_NO_SIGN_REQUEST"] = "yes"
                    os.environ["AWS_ACCESS_KEY_ID"] = "fake"
                    os.environ["AWS_SECRET_ACCESS_KEY"] = "fake"
                else:
                    unsigned = False
                    # delete env variable
                    del os.environ["AWS_NO_SIGN_REQUEST"]
            else:
                unsigned = False
                if log:
                    log.warning("AWS_NO_SIGN_REQUEST is not set. " +
                                "The default behaviour has recently changed to False (i.e. signed requests) " +
                                "Please explicitly set $AWS_NO_SIGN_REQUEST to 'no' for unsigned requests.")
            env_requester_pays = os.environ.get("AWS_REQUEST_PAYER", "")
            requester_pays = False
            if env_requester_pays.lower() == "requester":
                requester_pays = True
            self.use_aws = True
            if log:
                if unsigned:
                    log.info("S3 access configured with unsigned requests")
                else:
                    log.info("S3 access configured with signed requests")
            self.unsigned = unsigned
            self.requester_pays = requester_pays
            self.renew_creds()

            if "AWS_S3_ENDPOINT" in os.environ and os.environ["AWS_S3_ENDPOINT"] == "":
                del os.environ["AWS_S3_ENDPOINT"]
        elif log:
            log.warning(
                "Environment variable $AWS_DEFAULT_REGION not set.  (This warning can be ignored if all data is stored locally.)")

    def _check_cred(self) -> None:
        if self.credentials and isinstance(self.credentials, RefreshableCredentials):
            if self.credentials.refresh_needed():
                self.renew_creds()
            elif self.log:
                # pylint: disable=protected-access
                self.log.info("Credentials look OK: %s seconds remaining", str(self.credentials._seconds_remaining()))
        elif self.log:
            self.log.debug("Credentials of type %s - NOT RENEWING", self.credentials.__class__.__name__)

    @classmethod
    def check_cred(cls):
        # pylint: disable=protected-access
        cls._instance._check_cred()

    def renew_creds(self) -> None:
        if self.use_aws:
            if self.log:
                self.log.info("Establishing/renewing credentials")
            self.credentials = configure_s3_access(aws_unsigned=self.unsigned,
                                                            requester_pays=self.requester_pays)
            if self.log:
                if isinstance(self.credentials, RefreshableCredentials):
                    # pylint: disable=protected-access
                    self.log.debug("%s seconds remaining", str(self.credentials._seconds_remaining()))


def initialise_aws_credentials(log: Logger | None = None) -> None:
    # pylint: disable=protected-access
    if CredentialManager._instance is None:
        _ = CredentialManager(log)


def parse_config_file(log: Logger | None = None) -> OWSConfig | None:
    # Cache a parsed config file object
    # (unless deferring to first request)
    cfg = None
    if not os.environ.get("DEFER_CFG_PARSE"):
        cfg = get_config()
    return cfg


def initialise_flask(name) -> Flask:
    app_path = os.path.dirname(os.path.abspath(__file__))
    app = Flask(name.split('.')[0], template_folder=os.path.join(app_path, 'templates'))
    return app

def pass_through(undecorated):
    def decorator(*args, **kwargs):
        return undecorated(*args, **kwargs)
    decorator.__name__ = undecorated.__name__
    return decorator

class FakeMetrics:
    def do_not_track(self):
        return pass_through
    def counter(self, *args, **kwargs):
        return pass_through
    def histogram(self, *args, **kwargs):
        return pass_through
    def gauge(self, *args, **kwargs):
        return pass_through
    def summary(self, *args, **kwargs):
        return pass_through


def initialise_prometheus(app, log: Logger | None = None):
    # Prometheus
    if os.environ.get("PROMETHEUS_MULTIPROC_DIR", False):
        from prometheus_flask_exporter.multiprocess import \
            GunicornInternalPrometheusMetrics
        metrics = GunicornInternalPrometheusMetrics(app, group_by="endpoint")
        if log:
            log.info("Prometheus metrics enabled")
        return metrics
    return FakeMetrics()

def proxy_fix(app, log=None):
    # Proxy Fix, to respect X-Forwarded-For headers
    if os.environ.get("PROXY_FIX", False):
        from werkzeug.middleware.proxy_fix import ProxyFix
        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_port=1)
        if log is not None:
            log.info("ProxyFix was enabled")
    return app

def request_extractor():
    qreq = request.args.get('request')
    return qreq

def initialise_babel(cfg, app) -> object | None:
    if cfg and cfg.internationalised:
        from flask_babel import Babel
        app.config["BABEL_TRANSLATION_DIRECTORIES"] = cfg.translations_dir

        def get_locale():
            return request.accept_languages.best_match(cfg.locales, default=cfg.locales[0])

        babel = Babel(app,
                      locale_selector=get_locale,
                      default_domain=cfg.message_domain,
                      configure_jinja=False
                      )
        return babel
    else:
        return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0




bands_sentinel = {
    "B01": ["band_01", "coastal_aerosol"],
    "B02": ["band_02", "blue"],
    "B03": ["band_03", "green"],
    "B04": ["band_04", "red"],
    "B05": ["band_05", "red_edge_1"],
    "B06": ["band_06", "red_edge_2"],
    "B07": ["band_07", "red_edge_3"],
    "B08": ["band_08", "nir", "nir_1"],
    "B8A": ["band_8a", "nir_narrow", "nir_2"],
    "B09": ["band_09", "water_vapour"],
    "B11": ["band_11", "swir_1", "swir_16"],
    "B12": ["band_12", "swir_2", "swir_22"],
    "AOT": ["aerosol_optical_thickness"],
    "WVP": ["scene_average_water_vapour"],
    "SCL": ["mask", "qa"],
}


reslim_srtm = {
    "wms": {
        "zoomed_out_fill_colour": [150, 180, 200, 160],
        "min_zoom_factor": 10.0,
        # "max_datasets": 16, # Defaults to no dataset limit
    },
    "wcs": {
        # "max_datasets": 16, # Defaults to no dataset limit
    },
}




bands_sentinel = {
    "B01": ["band_01", "coastal_aerosol"],
    "B02": ["band_02", "blue"],
    "B03": ["band_03", "green"],
    "B04": ["band_04", "red"],
    "B05": ["band_05", "red_edge_1"],
    "B06": ["band_06", "red_edge_2"],
    "B07": ["band_07", "red_edge_3"],
    "B08": ["band_08", "nir", "nir_1"],
    "B8A": ["band_8a", "nir_narrow", "nir_2"],
    "B09": ["band_09", "water_vapour"],
    "B11": ["band_11", "swir_1", "swir_16"],
    "B12": ["band_12", "swir_2", "swir_22"],
    "AOT": ["aerosol_optical_thickness"],
    "WVP": ["scene_average_water_vapour"],
    "SCL": ["mask", "qa"],
}


reslim_srtm = {
    "wms": {
        "zoomed_out_fill_colour": [150, 180, 200, 160],
        "min_zoom_factor": 10.0,
        # "max_datasets": 16, # Defaults to no dataset limit
    },
    "wcs": {
        # "max_datasets": 16, # Defaults to no dataset limit
    },
}


# Reusable Chunks 3. Legends
legend_idx_0_1_5ticks = {
    "begin": "0.0",
    "end": "1.0",
    "ticks_every": "0.2",
    "units": "unitless",
}

legend_idx_percentage_by_10 = {
    "begin": "0.0",
    "end": "1.0",
    "ticks_every": 0.1,
    "units": "%",
    "tick_labels": {
        "0.0": {"label": "0"},
        "0.2": {"label": "20"},
        "0.4": {"label": "40"},
        "0.6": {"label": "60"},
        "0.8": {"label": "80"},
        "1.0": {"label": "100"},
    },
}